import logging
import time

from kikyo import Kikyo, DataHub

from kikyopp.consumer.base import BaseConsumer

log = logging.getLogger(__name__)


class KikyoConsumer(BaseConsumer):

    def __init__(self, worker):
        self.worker = worker
        self.is_running = True
        self.consumer = None
        self._delay = self.worker.settings.get('consumer_delay')

    @property
    def kikyo(self) -> Kikyo:
        return self.worker.kikyo

    @property
    def worker_name(self) -> str:
        return self.worker.name

    def run(self, name):
        if self.consumer is None:
            datahub = self.kikyo.component(cls=DataHub)
            self.consumer = datahub.subscribe(
                name,
                subscription_name=f'kikyopp.{self.worker_name}',
                auto_ack=False,
            )

        while self.is_running:
            try:
                if not self.is_running:
                    break
                msg = self.consumer.receive()

                if self._delay is not None:
                    t = msg.publish_time.timestamp() + self._delay - time.time()
                    if t > 600:
                        log.warning(f'Need to wait {t} seconds')
                    if t > 1:
                        time.sleep(t)

                data = msg.value
                self.worker.process(data)
                if not self.worker.debug:
                    self.consumer.ack(msg)
            except Exception as e:
                log.error(f'Error when consume data: {e}')

    def stop(self):
        self.is_running = False



import base64
import io
import logging
import signal

import requests
import yaml

from kikyopp.utils.logging import configure_logger
from kikyopp.worker import BaseWorker

log = logging.getLogger(__name__)


def run_worker(worker_cls, config_url: str, debug=False):
    log_level = 'debug' if debug else 'info'
    configure_logger('kikyopp', log_level)

    try:
        conf = _configure_by_consul(config_url)
        conf['debug'] = debug
        worker = worker_cls(conf)
        _set_signal_handlers(worker)
        worker.start()
    except Exception as e:
        log.error('Failed to start worker: %s', e, exc_info=True)


def _configure_by_consul(config_url) -> dict:
    resp = requests.get(config_url)
    resp.raise_for_status()
    v = resp.json()[0]['Value']
    s = base64.b64decode(v)
    conf = yaml.safe_load(io.BytesIO(s))
    return conf


def _set_signal_handlers(worker: BaseWorker):
    def _exit(signum, frame):
        log.info('Received exit signal: %s', signum)
        worker.stop()

    signal.signal(signal.SIGINT, _exit)
    signal.signal(signal.SIGTERM, _exit)



import datetime as dt

from pydantic import BaseModel


class PreprocessResult(BaseModel):
    class Config:
        validate_assignment = True

    id: str = None
    data_source: str = None
    create_time: dt.datetime = None
    update_time: dt.datetime = None



import inspect

KIKYOPP_SOURCE = '__kikyopp_source__'

KIKYOPP_SINK = '__kikyopp_sink__'


def source(name: str):
    """
    标识数据源

    :param name: 数据源名称
    """

    def wrapper(func):
        setattr(func, KIKYOPP_SOURCE, name)
        return func

    return wrapper


def sink(name: str):
    """
    标识数据处理器

    :param name: 数据处理器名称
    """

    def wrapper(func):
        setattr(func, KIKYOPP_SINK, name)
        return func

    if inspect.isfunction(name):
        f = name
        name = None
        return wrapper(f)
    return wrapper



import inspect
import json
import logging
from collections import defaultdict
from functools import partial

import gevent
from kikyo import Kikyo, DataHub
from kikyo.settings import Settings

from kikyopp.annotation import KIKYOPP_SOURCE, KIKYOPP_SINK
from kikyopp.consumer.base import BaseConsumer
from kikyopp.consumer.kikyo import KikyoConsumer

log = logging.getLogger(__name__)


class BaseWorker:
    name: str
    kikyo: Kikyo
    consumer: BaseConsumer

    def __init__(self, settings: dict):
        self.settings = Settings(settings)

        log.info('settings: %s', json.dumps(self.settings, ensure_ascii=False, indent=4))

        assert self.name is not None
        self.debug = self.settings.get('debug', False)
        self.workers = self.settings.get('workers', 1)

        self._sink = None
        self._sink_method = None
        self._source = None
        self._source_method = None
        self._producer = None

        self._source = defaultdict(list)
        for name in dir(self):
            method = getattr(self, name)
            if inspect.ismethod(method):
                if hasattr(method, KIKYOPP_SOURCE):
                    self._source = getattr(method, KIKYOPP_SOURCE)
                    self._source_method = method
                if hasattr(method, KIKYOPP_SINK):
                    self._sink = getattr(method, KIKYOPP_SINK)
                    self._sink_method = method

        self.is_running = False

        self._init_components()
        if self._sink:
            self._producer = self.kikyo.component(cls=DataHub).create_producer(self._sink)

    def _init_components(self):
        from kikyo import configure_by_consul

        self.kikyo = configure_by_consul(self.settings['kikyo_config_url'])
        self.consumer = KikyoConsumer(self)

    def start(self):
        if self.is_running:
            return

        if not self._source:
            log.warning('No defined source')

        self.is_running = True

        jobs = []
        for i in range(self.workers):
            jobs.append(
                gevent.spawn(partial(self.consumer.run, self._source))
            )
        gevent.wait(jobs)

    def process(self, data: dict):
        try:
            res = self._source_method(data)
            if res:
                if self._sink:
                    if not isinstance(res, list):
                        res = [res]
                    for r in res:
                        self._producer.send(r)
                        self._sink_method(r)
        except Exception as e:
            log.error(f'Error occurred: {e}', exc_info=True)

    def stop(self):
        if not self.is_running:
            return
        self.is_running = False
        self.consumer.stop()



import datetime as dt
from typing import Union, List

from pydantic import BaseModel


class SearchInfo(BaseModel):
    class Config:
        validate_assignment = True

    A_title: str = None
    A_content: Union[str, List[str]] = None
    A_file_path_list: List[str] = []
    A_file_name_list: List[str] = []
    A_label_list: List[str] = []
    A_create_time: dt.datetime = None
